@@ -6,7 +6,7 @@ import os
 
 from typing import DefaultDict, Dict, List, Set, Tuple, Optional, Literal, TypedDict
 
-from .tables import SUFFIX_TYPES, UNIVERSAL_FEATURES, XPOS_TO_UPOS
+from .tables import SUFFIX_TYPES, UNIVERSAL_FEATURES, XPOS_TO_UPOS, PREFIX_TABLE, Person, Number
 
 Xpos = Literal["VS", "VT", "VI", "VA", "V?", "NL", "NB", "PRON", "NUM", "N", "ADV", "EXCL", "CONJ", "QUES", "UNK"]
 
@@ -168,11 +168,58 @@ def _next_morphem(parsed: List[str], i: int) -> Tuple[int, Optional[str]]:
     
     return oi, None
 
+LETTER = re.compile(r"(ch|gh|ng|tlh|[abDeHIjlmnopqQrStuvwy'])")
+
+def split_to_letters(word: str) -> List[str]:
+    """
+    Splits the given word to Klingon letters. The word must not contain spaces or foreign letters.
+    """
+    letters = []
+    while word:
+        if m := re.match(LETTER, word):
+            letter = m.group(1)
+            word = word[len(letter):]
+            letters += [letter]
+
+        else:
+            word = word[1:]
+
+    return letters
+
+def split_to_syllables(word: str) -> List[str]:
+    """
+    Splits the given word to syllables. The word msut not contain spaces or foreign words and it should follow the Klingon phonotactics.
+    """
+    letters = split_to_letters(word)
+    syllables = [""]
+
+    for prev, cur, succ in zip([""]+letters, letters, letters[1:]+[""]):
+        if syllables[-1] and \
+            cur not in "aeIou" and \
+            (succ and succ in "aeIou" or \
+                prev and prev not in "aeIou" and \
+                not (prev+cur) in {"w'", "y'", "rgh"}):
+            syllables.append(cur)
+
+        else:
+            syllables[-1] += cur
+    
+    return syllables
+
+class SyntaxInfo(TypedDict, total=False):
+    ROLE: Literal["NP", "VP", "OTHER"]
+    OBJECT_PERSON: Set[Person]
+    OBJECT_NUMBER: Optional[Number]
+    SUBJECT_PERSON: Set[Person]
+    SUBJECT_NUMBER: Optional[Number]
+    PLURAL: bool
+
 class _AnalysisRequired(TypedDict):
     WORD: str
     LEMMA: str
     POS: str
     XPOS: Xpos
+    BOQWIZ_POS: str
     PARTS: List[str]
 
 class Analysis(_AnalysisRequired, total=False):
@@ -180,6 +227,7 @@ class Analysis(_AnalysisRequired, total=False):
     SUFFIX: Dict[str, str]
     XPOS_GSUFF: str
     UNGRAMMATICAL: bool
+    SYNTAX_INFO: SyntaxInfo
 
 def _analyze_word_with_pos(ans: List[Analysis], start_pos: str, regex: re.Pattern[str], lemma_idx: int, word: str, infl_pos:str=None, lemma_pred=lambda l: True):
     if m := regex.fullmatch(word):
@@ -231,6 +279,7 @@ def _analyze_word_with_pos(ans: List[Analysis], start_pos: str, regex: re.Patter
                         #new_obj["BOQWIZ"] = entry
 
                         new_obj["XPOS"] = _get_xpos(entry)
+                        new_obj["BOQWIZ_POS"] = entry["part_of_speech"]
 
                         if not lemma_pred(entry):
                             continue
@@ -248,6 +297,7 @@ def _analyze_word_with_pos(ans: List[Analysis], start_pos: str, regex: re.Patter
             "WORD": word,
             "POS": start_pos.upper(),
             "XPOS": "UNK",
+            "BOQWIZ_POS": "?",
             "PARTS": [],
             "LEMMA": "",
         }
@@ -268,7 +318,7 @@ GENDERED_SUFFIXES = {
     "other": set(),
 }
 
-def analyze(word: str) -> List[Analysis]:
+def analyze(word: str, include_syntactical_info=False) -> List[Analysis]:
     """
     Given a word, returns a list of possible analyses.
 
@@ -299,6 +349,7 @@ def analyze(word: str) -> List[Analysis]:
                 "LEMMA": entry["entry_name"],
                 "POS": "OTHER",
                 "XPOS": _get_xpos(entry),
+                "BOQWIZ_POS": entry["part_of_speech"],
                 "PARTS": [ entry["entry_name"] + ":" + entry["part_of_speech"] ],
             })
     
@@ -327,11 +378,66 @@ def analyze(word: str) -> List[Analysis]:
                         analysis["UNGRAMMATICAL"] = True
         
         if "-vIS:v:suff" in analysis["PARTS"] and \
-            "-taH:v:suff" not in analysis["PARTS"] and "-lI':v:suff" not in analysis["PARTS"]:
+            "-taH:v:suff" not in analysis["PARTS"]:
             analysis["UNGRAMMATICAL"] = True
         
         if "-lu':v:suff" in analysis["PARTS"] and analysis.get("PREFIX", "") not in {"vI-", "Da-", "wI-", "bo-", "", "lu-"}:
             analysis["UNGRAMMATICAL"] = True
+        
+        if analysis["XPOS"] in {"VS", "VI"} and analysis.get("PREFIX", "") not in {"yI-", "pe-", "jI-", "bI-", "ma-", "Su-", ""} and "-moH:v:suff" not in analysis["PARTS"]:
+            analysis["UNGRAMMATICAL"] = True
+        
+        if "-ghach:v:suff" in analysis["PARTS"] and analysis["PARTS"].index("-ghach:v:suff") - analysis["PARTS"].index(analysis["LEMMA"] + ":" + analysis["BOQWIZ_POS"]) < 2:
+            analysis["UNGRAMMATICAL"] = True
+        
+        # Add extra information regarding the words rule in the syntax
+        if include_syntactical_info:
+            info: SyntaxInfo = {}
+            analysis["SYNTAX_INFO"] = info
+            if analysis["POS"] == "N":
+                info["ROLE"] = "NP"
+            
+            elif analysis["POS"] == "V" and analysis.get("SUFFIX", {}).get("V9", None) in {"-wI'", "-ghach"}:
+                info["ROLE"] = "NP"
+            
+            elif analysis["POS"] == "V":
+                info["ROLE"] = "VP"
+            
+            else:
+                info["ROLE"] = "OTHER"
+            
+            if info["ROLE"] == "VP":
+                voice = "NP" if "-lu':v:suff" in analysis["PARTS"] else "P"
+                subj_person: Set[Person]
+                subj_number: Optional[Number]
+                obj_person: Set[Person]
+                obj_number: Optional[Number]
+                if "PREFIX" in analysis or voice == "NP":
+                    subj_person, subj_number, obj_person, obj_number = PREFIX_TABLE[(analysis.get("PREFIX", "-"), voice)]
+                
+                elif analysis["XPOS"] in {"VS", "VI"}:
+                    subj_person = {3}
+                    subj_number = None
+                    obj_person = {0}
+                    obj_number = None
+
+                else:
+                    subj_person = {3}
+                    subj_number = None
+                    obj_person = {0, 3}
+                    obj_number = None
+                
+                info["SUBJECT_PERSON"] = subj_person
+                info["SUBJECT_NUMBER"] = subj_number
+                info["OBJECT_PERSON"] = obj_person
+                info["OBJECT_NUMBER"] = obj_number
+            
+            elif info["ROLE"] == "NP":
+                if "inhps" in analysis["BOQWIZ_POS"] or "inhpl" in analysis["BOQWIZ_POS"]:
+                    info["PLURAL"] = False
+
+                if analysis.get("SUFFIX", {}).get("N2", None) in {"-pu'", "-Du'", "-mey"}:
+                    info["PLURAL"] = True
 
     return ans
 
@@ -378,18 +484,19 @@ def tokenize(sentence: str) -> List[Tuple[TokenType, str]]:
     """
     Tokenizes the given text and returns a list of tuples with form `(type, value)` where type is one of `"WORD"`, `"SPACE"`, `"PUNCT"`.
     """
-    tokens: List[Tuple[TokenType, str]] = [("WORD", "")]
-    for char in sentence:
-        if re.match(r"[a-zA-Z'0-9]", char):
-            tokens[-1] = ("WORD", tokens[-1][1] + char)
+    tokens: List[Tuple[TokenType, str]] = []
+    for matches in re.findall(r"([a-zA-Z'0-9]+|\s+|(.)\2*)", sentence):
+        token = matches[0]
+        if re.fullmatch(r"[a-zA-Z'0-9]+", token):
+            tokens.append(("WORD", token))
         
-        elif re.match(r"\s", char):
-            tokens += [("SPACE", char), ("WORD", "")]
+        elif re.fullmatch(r"\s+", token):
+            tokens.append(("SPACE", token))
         
-        else:
-            tokens += [("PUNCT", char), ("WORD", "")]
+        elif token:
+            tokens.append(("PUNCT", token))
     
-    return [token for token in tokens if token[1]]
+    return tokens
 
 def text_to_conllu_without_tagger(text: str) -> str:
     """
@@ -417,4 +524,4 @@ def text_to_conllu_without_tagger(text: str) -> str:
             ans.append("\t".join(_word_to_conllu(i, token, analyses)))
             i += 1
     
-    return "\n".join(ans)
\ No newline at end of file
+    return "\n".join(ans)
